.set noreorder

#include "pspimport.s"

IMPORT_START "sceAtrac3plus",0x00090000
IMPORT_FUNC  "sceAtrac3plus",0xD1F59FDB,sceAtracStartEntry
IMPORT_FUNC  "sceAtrac3plus",0xD5C28CC0,sceAtracEndEntry
IMPORT_FUNC  "sceAtrac3plus",0x780F88D1,sceAtracGetAtracID
IMPORT_FUNC  "sceAtrac3plus",0x61EB33F5,sceAtracReleaseAtracID
IMPORT_FUNC  "sceAtrac3plus",0x0E2A73AB,sceAtracSetData
IMPORT_FUNC  "sceAtrac3plus",0x3F6E26B5,sceAtracSetHalfwayBuffer
IMPORT_FUNC  "sceAtrac3plus",0x7A20E7AF,sceAtracSetDataAndGetID
IMPORT_FUNC  "sceAtrac3plus",0x0FAE370E,sceAtracSetHalfwayBufferAndGetID
IMPORT_FUNC  "sceAtrac3plus",0x6A8C3CD5,sceAtracDecodeData
IMPORT_FUNC  "sceAtrac3plus",0x9AE849A7,sceAtracGetRemainFrame
IMPORT_FUNC  "sceAtrac3plus",0x5D268707,sceAtracGetStreamDataInfo
IMPORT_FUNC  "sceAtrac3plus",0x7DB31251,sceAtracAddStreamData
IMPORT_FUNC  "sceAtrac3plus",0x83E85EA0,sceAtracGetSecondBufferInfo
IMPORT_FUNC  "sceAtrac3plus",0x83BF7AFD,sceAtracSetSecondBuffer
IMPORT_FUNC  "sceAtrac3plus",0xE23E3A35,sceAtracGetNextDecodePosition
IMPORT_FUNC  "sceAtrac3plus",0xA2BBA8BE,sceAtracGetSoundSample
IMPORT_FUNC  "sceAtrac3plus",0x31668BAA,sceAtracGetChannel
IMPORT_FUNC  "sceAtrac3plus",0xD6A5F2F7,sceAtracGetMaxSample
IMPORT_FUNC  "sceAtrac3plus",0x36FAABFB,sceAtracGetNextSample
IMPORT_FUNC  "sceAtrac3plus",0xA554A158,sceAtracGetBitrate
IMPORT_FUNC  "sceAtrac3plus",0xFAA4F89B,sceAtracGetLoopStatus
IMPORT_FUNC  "sceAtrac3plus",0x868120B5,sceAtracSetLoopNum
IMPORT_FUNC  "sceAtrac3plus",0xCA3CA3D2,sceAtracGetBufferInfoForReseting
IMPORT_FUNC  "sceAtrac3plus",0x644E5607,sceAtracResetPlayPosition
IMPORT_FUNC  "sceAtrac3plus",0xE88F759B,sceAtracGetInternalErrorInfo
IMPORT_FUNC  "sceAtrac3plus",0x1575D64B,sceAtracLowLevelInitDecoder
IMPORT_FUNC  "sceAtrac3plus",0x0C116E1B,sceAtracLowLevelDecode
IMPORT_FUNC  "sceAtrac3plus",0x132f1eca,sceAtracReinit
IMPORT_FUNC  "sceAtrac3plus",0xeca32a99,sceAtracIsSecondBufferNeeded
IMPORT_FUNC  "sceAtrac3plus",0x2DD3E298,sceAtracGetBufferInfoForResetting
IMPORT_FUNC  "sceAtrac3plus",0x5CF9D852,sceAtracSetMOutHalfwayBuffer
IMPORT_FUNC  "sceAtrac3plus",0xF6837A1A,sceAtracSetMOutData
IMPORT_FUNC  "sceAtrac3plus",0x472E3825,sceAtracSetMOutDataAndGetID
IMPORT_FUNC  "sceAtrac3plus",0x9CD7DE03,sceAtracSetMOutHalfwayBufferAndGetID
IMPORT_FUNC  "sceAtrac3plus",0xB3B5D042,sceAtracGetOutputChannel
IMPORT_FUNC  "sceAtrac3plus",0x5622B7C1,sceAtracSetAA3DataAndGetID
IMPORT_FUNC  "sceAtrac3plus",0x5DD66588,sceAtracSetAA3HalfwayBufferAndGetID
IMPORT_FUNC  "sceAtrac3plus",0x231FC6B7,_sceAtracGetContextAddress