.set noreorder

#include "pspimport.s"

IMPORT_START "ThreadManForUser",0x40010000
IMPORT_FUNC	"ThreadManForUser",0xB7D098C6,sceKernelCreateMutex
IMPORT_FUNC	"ThreadManForUser",0xF8170FBE,sceKernelDeleteMutex
IMPORT_FUNC	"ThreadManForUser",0x6B30100F,sceKernelUnlockMutex
IMPORT_FUNC	"ThreadManForUser",0xB011B11F,sceKernelLockMutex
IMPORT_FUNC	"ThreadManForUser",0x5bf4dd27,sceKernelLockMutexCB
IMPORT_FUNC	"ThreadManForUser",0x0DDCD2C9,sceKernelTryLockMutex
IMPORT_FUNC	"ThreadManForUser",0x60107536,sceKernelDeleteLwMutex
IMPORT_FUNC	"ThreadManForUser",0x19CFF145,sceKernelCreateLwMutex
IMPORT_FUNC	"ThreadManForUser",0x4C145944,sceKernelReferLwMutexStatusByID

IMPORT_FUNC	"ThreadManForUser",0x6E9EA350,_sceKernelReturnFromCallback
IMPORT_FUNC	"ThreadManForUser",0x0C106E53,sceKernelRegisterThreadEventHandler
IMPORT_FUNC	"ThreadManForUser",0x72F3C145,sceKernelReleaseThreadEventHandler
IMPORT_FUNC	"ThreadManForUser",0x369EEB6B,sceKernelReferThreadEventHandlerStatus
IMPORT_FUNC	"ThreadManForUser",0xE81CAF8F,sceKernelCreateCallback
IMPORT_FUNC	"ThreadManForUser",0xEDBA5844,sceKernelDeleteCallback
IMPORT_FUNC	"ThreadManForUser",0xC11BA8C4,sceKernelNotifyCallback
IMPORT_FUNC	"ThreadManForUser",0xBA4051D6,sceKernelCancelCallback
IMPORT_FUNC	"ThreadManForUser",0x2A3D44FF,sceKernelGetCallbackCount
IMPORT_FUNC	"ThreadManForUser",0x349D6D6C,sceKernelCheckCallback
IMPORT_FUNC	"ThreadManForUser",0x730ED8BC,sceKernelReferCallbackStatus
IMPORT_FUNC	"ThreadManForUser",0x9ACE131E,sceKernelSleepThread
IMPORT_FUNC	"ThreadManForUser",0x82826F70,sceKernelSleepThreadCB
IMPORT_FUNC	"ThreadManForUser",0xD59EAD2F,sceKernelWakeupThread
IMPORT_FUNC	"ThreadManForUser",0xFCCFAD26,sceKernelCancelWakeupThread
IMPORT_FUNC	"ThreadManForUser",0x9944F31F,sceKernelSuspendThread
IMPORT_FUNC	"ThreadManForUser",0x75156E8F,sceKernelResumeThread
IMPORT_FUNC	"ThreadManForUser",0x278C0DF5,sceKernelWaitThreadEnd
IMPORT_FUNC	"ThreadManForUser",0x840E8133,sceKernelWaitThreadEndCB
IMPORT_FUNC	"ThreadManForUser",0xCEADEB47,sceKernelDelayThread
IMPORT_FUNC	"ThreadManForUser",0x68DA9E36,sceKernelDelayThreadCB
IMPORT_FUNC	"ThreadManForUser",0xBD123D9E,sceKernelDelaySysClockThread
IMPORT_FUNC	"ThreadManForUser",0x1181E963,sceKernelDelaySysClockThreadCB
IMPORT_FUNC	"ThreadManForUser",0xD6DA4BA1,sceKernelCreateSema
IMPORT_FUNC	"ThreadManForUser",0x28B6489C,sceKernelDeleteSema
IMPORT_FUNC	"ThreadManForUser",0x3F53E640,sceKernelSignalSema
IMPORT_FUNC	"ThreadManForUser",0x4E3A1105,sceKernelWaitSema
IMPORT_FUNC	"ThreadManForUser",0x6D212BAC,sceKernelWaitSemaCB
IMPORT_FUNC	"ThreadManForUser",0x58B1F937,sceKernelPollSema
IMPORT_FUNC	"ThreadManForUser",0x8FFDF9A2,sceKernelCancelSema
IMPORT_FUNC	"ThreadManForUser",0xBC6FEBC5,sceKernelReferSemaStatus
IMPORT_FUNC	"ThreadManForUser",0x55C20A00,sceKernelCreateEventFlag
IMPORT_FUNC	"ThreadManForUser",0xEF9E4C70,sceKernelDeleteEventFlag
IMPORT_FUNC	"ThreadManForUser",0x1FB15A32,sceKernelSetEventFlag
IMPORT_FUNC	"ThreadManForUser",0x812346E4,sceKernelClearEventFlag
IMPORT_FUNC	"ThreadManForUser",0x402FCF22,sceKernelWaitEventFlag
IMPORT_FUNC	"ThreadManForUser",0x328C546A,sceKernelWaitEventFlagCB
IMPORT_FUNC	"ThreadManForUser",0x30FD48F0,sceKernelPollEventFlag
IMPORT_FUNC	"ThreadManForUser",0xCD203292,sceKernelCancelEventFlag
IMPORT_FUNC	"ThreadManForUser",0xA66B0120,sceKernelReferEventFlagStatus
IMPORT_FUNC	"ThreadManForUser",0x8125221D,sceKernelCreateMbx
IMPORT_FUNC	"ThreadManForUser",0x86255ADA,sceKernelDeleteMbx
IMPORT_FUNC	"ThreadManForUser",0xE9B3061E,sceKernelSendMbx
IMPORT_FUNC	"ThreadManForUser",0x18260574,sceKernelReceiveMbx
IMPORT_FUNC	"ThreadManForUser",0xF3986382,sceKernelReceiveMbxCB
IMPORT_FUNC	"ThreadManForUser",0x0D81716A,sceKernelPollMbx
IMPORT_FUNC	"ThreadManForUser",0x87D4DD36,sceKernelCancelReceiveMbx
IMPORT_FUNC	"ThreadManForUser",0xA8E8C846,sceKernelReferMbxStatus
IMPORT_FUNC	"ThreadManForUser",0x7C0DC2A0,sceKernelCreateMsgPipe
IMPORT_FUNC	"ThreadManForUser",0xF0B7DA1C,sceKernelDeleteMsgPipe
IMPORT_FUNC	"ThreadManForUser",0x876DBFAD,sceKernelSendMsgPipe
IMPORT_FUNC	"ThreadManForUser",0x7C41F2C2,sceKernelSendMsgPipeCB
IMPORT_FUNC	"ThreadManForUser",0x884C9F90,sceKernelTrySendMsgPipe
IMPORT_FUNC	"ThreadManForUser",0x74829B76,sceKernelReceiveMsgPipe
IMPORT_FUNC	"ThreadManForUser",0xFBFA697D,sceKernelReceiveMsgPipeCB
IMPORT_FUNC	"ThreadManForUser",0xDF52098F,sceKernelTryReceiveMsgPipe
IMPORT_FUNC	"ThreadManForUser",0x349B864D,sceKernelCancelMsgPipe
IMPORT_FUNC	"ThreadManForUser",0x33BE4024,sceKernelReferMsgPipeStatus
IMPORT_FUNC	"ThreadManForUser",0x56C039B5,sceKernelCreateVpl
IMPORT_FUNC	"ThreadManForUser",0x89B3D48C,sceKernelDeleteVpl
IMPORT_FUNC	"ThreadManForUser",0xBED27435,sceKernelAllocateVpl
IMPORT_FUNC	"ThreadManForUser",0xEC0A693F,sceKernelAllocateVplCB
IMPORT_FUNC	"ThreadManForUser",0xAF36D708,sceKernelTryAllocateVpl
IMPORT_FUNC	"ThreadManForUser",0xB736E9FF,sceKernelFreeVpl
IMPORT_FUNC	"ThreadManForUser",0x1D371B8A,sceKernelCancelVpl
IMPORT_FUNC	"ThreadManForUser",0x39810265,sceKernelReferVplStatus
IMPORT_FUNC	"ThreadManForUser",0xC07BB470,sceKernelCreateFpl
IMPORT_FUNC	"ThreadManForUser",0xED1410E0,sceKernelDeleteFpl
IMPORT_FUNC	"ThreadManForUser",0xD979E9BF,sceKernelAllocateFpl
IMPORT_FUNC	"ThreadManForUser",0xE7282CB6,sceKernelAllocateFplCB
IMPORT_FUNC	"ThreadManForUser",0x623AE665,sceKernelTryAllocateFpl
IMPORT_FUNC	"ThreadManForUser",0xF6414A71,sceKernelFreeFpl
IMPORT_FUNC	"ThreadManForUser",0xA8AA591F,sceKernelCancelFpl
IMPORT_FUNC	"ThreadManForUser",0xD8199E4C,sceKernelReferFplStatus
IMPORT_FUNC	"ThreadManForUser",0x0E927AED,_sceKernelReturnFromTimerHandler
IMPORT_FUNC	"ThreadManForUser",0x110DEC9A,sceKernelUSec2SysClock
IMPORT_FUNC	"ThreadManForUser",0xC8CD158C,sceKernelUSec2SysClockWide
IMPORT_FUNC	"ThreadManForUser",0xBA6B92E2,sceKernelSysClock2USec
IMPORT_FUNC	"ThreadManForUser",0xE1619D7C,sceKernelSysClock2USecWide
IMPORT_FUNC	"ThreadManForUser",0xDB738F35,sceKernelGetSystemTime
IMPORT_FUNC	"ThreadManForUser",0x82BC5777,sceKernelGetSystemTimeWide
IMPORT_FUNC	"ThreadManForUser",0x369ED59D,sceKernelGetSystemTimeLow
IMPORT_FUNC	"ThreadManForUser",0x6652B8CA,sceKernelSetAlarm
IMPORT_FUNC	"ThreadManForUser",0xB2C25152,sceKernelSetSysClockAlarm
IMPORT_FUNC	"ThreadManForUser",0x7E65B999,sceKernelCancelAlarm
IMPORT_FUNC	"ThreadManForUser",0xDAA3F564,sceKernelReferAlarmStatus
IMPORT_FUNC	"ThreadManForUser",0x20FFF560,sceKernelCreateVTimer
IMPORT_FUNC	"ThreadManForUser",0x328F9E52,sceKernelDeleteVTimer
IMPORT_FUNC	"ThreadManForUser",0xB3A59970,sceKernelGetVTimerBase
IMPORT_FUNC	"ThreadManForUser",0xB7C18B77,sceKernelGetVTimerBaseWide
IMPORT_FUNC	"ThreadManForUser",0x034A921F,sceKernelGetVTimerTime
IMPORT_FUNC	"ThreadManForUser",0xC0B3FFD2,sceKernelGetVTimerTimeWide
IMPORT_FUNC	"ThreadManForUser",0x542AD630,sceKernelSetVTimerTime
IMPORT_FUNC	"ThreadManForUser",0xFB6425C3,sceKernelSetVTimerTimeWide
IMPORT_FUNC	"ThreadManForUser",0xC68D9437,sceKernelStartVTimer
IMPORT_FUNC	"ThreadManForUser",0xD0AEEE87,sceKernelStopVTimer
IMPORT_FUNC	"ThreadManForUser",0xD8B299AE,sceKernelSetVTimerHandler
IMPORT_FUNC	"ThreadManForUser",0x53B00E9A,sceKernelSetVTimerHandlerWide
IMPORT_FUNC	"ThreadManForUser",0xD2D615EF,sceKernelCancelVTimerHandler
IMPORT_FUNC	"ThreadManForUser",0x5F32BEAA,sceKernelReferVTimerStatus
IMPORT_FUNC	"ThreadManForUser",0x446D8DE6,sceKernelCreateThread
IMPORT_FUNC	"ThreadManForUser",0x9FA03CD3,sceKernelDeleteThread
IMPORT_FUNC	"ThreadManForUser",0xF475845D,sceKernelStartThread
IMPORT_FUNC	"ThreadManForUser",0x532A522E,_sceKernelExitThread
IMPORT_FUNC	"ThreadManForUser",0xAA73C935,sceKernelExitThread
IMPORT_FUNC	"ThreadManForUser",0x809CE29B,sceKernelExitDeleteThread
IMPORT_FUNC	"ThreadManForUser",0x616403BA,sceKernelTerminateThread
IMPORT_FUNC	"ThreadManForUser",0x383F7BCC,sceKernelTerminateDeleteThread
IMPORT_FUNC	"ThreadManForUser",0x3AD58B8C,sceKernelSuspendDispatchThread
IMPORT_FUNC	"ThreadManForUser",0x27E22EC2,sceKernelResumeDispatchThread
IMPORT_FUNC	"ThreadManForUser",0xEA748E31,sceKernelChangeCurrentThreadAttr
IMPORT_FUNC	"ThreadManForUser",0x71BC9871,sceKernelChangeThreadPriority
IMPORT_FUNC	"ThreadManForUser",0x912354A7,sceKernelRotateThreadReadyQueue
IMPORT_FUNC	"ThreadManForUser",0x2C34E053,sceKernelReleaseWaitThread
IMPORT_FUNC	"ThreadManForUser",0x293B45B8,sceKernelGetThreadId
IMPORT_FUNC	"ThreadManForUser",0x94AA61EE,sceKernelGetThreadCurrentPriority
IMPORT_FUNC	"ThreadManForUser",0x3B183E26,sceKernelGetThreadExitStatus
IMPORT_FUNC	"ThreadManForUser",0xD13BDE95,sceKernelCheckThreadStack
IMPORT_FUNC	"ThreadManForUser",0x52089CA1,sceKernelGetThreadStackFreeSize
IMPORT_FUNC	"ThreadManForUser",0x17C1684E,sceKernelReferThreadStatus
IMPORT_FUNC	"ThreadManForUser",0xFFC36A14,sceKernelReferThreadRunStatus
IMPORT_FUNC	"ThreadManForUser",0x627E6F3A,sceKernelReferSystemStatus
IMPORT_FUNC	"ThreadManForUser",0x94416130,sceKernelGetThreadmanIdList
IMPORT_FUNC	"ThreadManForUser",0x57CF62DD,sceKernelGetThreadmanIdType
IMPORT_FUNC	"ThreadManForUser",0x64D4540E,sceKernelReferThreadProfiler
IMPORT_FUNC	"ThreadManForUser",0x8218B4DD,sceKernelReferGlobalProfiler

IMPORT_START "Kernel_Library",0x00010000
IMPORT_FUNC "Kernel_Library",0x092968F4,sceKernelCpuSuspendIntr
IMPORT_FUNC "Kernel_Library",0x5F10D406,sceKernelCpuResumeIntr
IMPORT_FUNC "Kernel_Library",0x3B84732D,sceKernelCpuResumeIntrWithSync
IMPORT_FUNC "Kernel_Library",0x47A0B729,sceKernelIsCpuIntrSuspended
IMPORT_FUNC "Kernel_Library",0xB55249D2,sceKernelIsCpuIntrEnable

IMPORT_FUNC "Kernel_Library",0xa089eca4,sceKernelMemset
IMPORT_FUNC "Kernel_Library",0xDC692EE3,sceKernelTryLockLwMutex
IMPORT_FUNC "Kernel_Library",0x37431849,sceKernelTryLockLwMutex_600
IMPORT_FUNC "Kernel_Library",0xbea46419,sceKernelLockLwMutex
IMPORT_FUNC "Kernel_Library",0x1FC64E09,sceKernelLockLwMutexCB
IMPORT_FUNC "Kernel_Library",0x15b6446b,sceKernelUnlockLwMutex
IMPORT_FUNC "Kernel_Library",0xC1734599,sceKernelReferLwMutexStatus

IMPORT_FUNC "Kernel_Library",0x1839852A,sce_paf_private_memcpy
IMPORT_FUNC "Kernel_Library",0xA089ECA4,sce_paf_private_memset
